<?php
/**
 * Discount amount calculator
 */
class Calculator_Discount
{
	protected static $instance;

	protected $settings;
	protected $orderRepository;
	protected $discountRepository;

    /**
     * Class constructor
     *
     * @param $settings
     * @param DataAccess_OrderRepositoryInterface $orderRepository
     * @param DataAccess_DiscountRepositoryInterface $discountRepository
     */
    public function __construct(&$settings, DataAccess_OrderRepositoryInterface $orderRepository, DataAccess_DiscountRepositoryInterface $discountRepository)
	{
		$this->settings = $settings;
		$this->orderRepository = $orderRepository;
		$this->discountRepository = $discountRepository;
	}

	/**
	 * Reset discounts
	 *
	 * @param ORDER $order
	 */
	public function reset(ORDER $order)
	{
		$order->setDiscountAmount(0);
		$order->setDiscountValue(0);
		$order->setDiscountType('none');

		foreach ($order->lineItems as $item)
		{
			/** @var Model_LineItem $item */
			$item->setDiscountAmount(0);
			$item->setPromoDiscountAmount(0);
		}

		$this->orderRepository->resetLineItemDiscounts($order->getId());
	}

	/**
	 * Calculate discount
	 *
	 * @param ORDER $order
	 * @param bool $reset
	 *
	 * @return string
	 */
	public function calculate(ORDER $order, $reset = true)
	{
		$subtotal = $order->calculateSubtotal(false, true);

		$discountAmount = 0;

		if ($reset)
		{
			$this->reset($order);

			if ($order->getUserLevel() < 1)
			{
				if ($this->settings["DiscountsActive"] == "YES")
				{
					$discount = $this->discountRepository->getActiveDiscount($subtotal);

					if ($discount)
					{
						$order->setDiscountType($discount["type"]);
						$order->setDiscountValue($discount["discount"]);

						$dp = $this->getDiscountPercentage($subtotal, $order->getDiscountType(), $order->getDiscountValue());

						foreach ($order->lineItems as $item)
						{
							/* @var $item Model_LineItem */
							if ($item->getProductId() == 'gift_certificate')
								continue;

							$lineSubTotal = $item->getSubtotal();

							$lineDiscountAmount = PriceHelper::round($lineSubTotal * $dp / 100);
							$item->setDiscountAmount($lineDiscountAmount);
							$discountAmount += $lineDiscountAmount;

							$this->orderRepository->updateLineItemDiscount($order->getId(), $item->getId(), $lineDiscountAmount, $item->getPromoDiscountAmount());
						}
					}
				}
			}
		}
		else
		{
			foreach ($order->lineItems as $item)
			{
				/* @var $item Model_LineItem */
				$discountAmount += $item->getDiscountAmount();
			}
		}

		$order->setDiscountAmount($discountAmount);

		return $order->getDiscountAmount();
	}

	/**
	 * Get discount percentage
	 *
	 * @param $subtotalAmount
	 * @param $discountType
	 * @param $discountValue
	 *
	 * @return float|int
	 */
	protected function getDiscountPercentage($subtotalAmount, $discountType, $discountValue)
	{
		//determinate discount percent
		$dp = 0;

		if ($subtotalAmount > 0)
		{
			if ($discountType == 'percent')
			{
				$dp = $discountValue;
			}
			else if ($discountType == 'amount')
			{
				$dp = $discountValue / ($subtotalAmount / 100);
			}
		}

		return $dp > 100 ? 100 : $dp;
	}

	/**
	 * Get class instance
	 *
	 * @param $settings
	 * @param bool $orderRepository
	 * @param bool $discountRepository
	 *
	 * @return Calculator_Discount
	 */
	// @codeCoverageIgnoreStart
	public static function getInstance(&$settings, $orderRepository = false, $discountRepository = false)
	{
		if (is_null(self::$instance))
		{
			if (!$orderRepository)
			{
				$orderRepository = new DataAccess_OrderRepository();
			}
			if (!$discountRepository)
			{
				$discountRepository = new DataAccess_DiscountRepository();
			}

			self::$instance = new self($settings, $orderRepository, $discountRepository);
		}

		return self::$instance;
	}

	/**
	 * Get instance
	 *
	 * @param Calculator_Discount $instance
	 */
	public static function setInstance(Calculator_Discount $instance)
	{
		self::$instance = $instance;
	}
	// @codeCoverageIgnoreEnd
}