<?php
/**
 * Calculate promo discount
 */
class Calculator_PromoDiscount
{
	protected static $instance;

	protected $orderRepository;
	protected $promoCodeRepository;
	protected $settings;

	/**
	 * Class constructor
	 *
	 * @param DataAccess_OrderRepositoryInterface $orderRepository
	 * @param DataAccess_PromoCodeRepositoryInterface $promoCodeRepository
	 *
	 * @param $settings
	 */
	public function __construct(DataAccess_OrderRepositoryInterface $orderRepository, DataAccess_PromoCodeRepositoryInterface $promoCodeRepository, &$settings)
	{
		$this->orderRepository = $orderRepository;
		$this->promoCodeRepository = $promoCodeRepository;
		$this->settings = $settings;
	}

	/**
	 * Reset promo code data
	 *
	 * @param ORDER $order
	 * @param bool $justAmount
	 */
	public function reset(ORDER $order, $justAmount = false)
	{
		if ($justAmount)
		{
			$order->setPromoDiscountAmount(0);

			foreach ($order->lineItems as $item)
			{
				/** @var Model_LineItem $item */
				$item->setPromoDiscountAmount(0);
				$this->orderRepository->updateLineItemDiscount($order->getId(), $item->getId(), $item->getDiscountAmount(), 0);
			}
		}
		else
		{
			$this->setOrderPromoData($order, 'Global', 0, 'none', 0, 0);

			foreach ($order->lineItems as $item)
			{
				/** @var Model_LineItem $item */
				$item->setPromoDiscountAmount(0);
				$this->orderRepository->updateLineItemDiscount($order->getId(), $item->getId(), $item->getDiscountAmount(), 0);
			}
		}
	}

	/**
	 * Calculate
	 * 
	 * @param ORDER $order
	 * @param boolean $reset
	 *
	 * @return mixed
	 */
	public function calculate(ORDER $order, $reset = true)
	{
		// Calculate only if promo enabled and was campaign was set
		if ($this->settings['DiscountsPromo'] == 'YES' && $order->getPromoCampaignId() > 0)
		{
			$subtotalAmount = $order->calculateSubtotal(false, true);

			if ($reset)
			{
				// reread promo data from data storage
				$promoData = $this->promoCodeRepository->getPromoCodeByPromoCampaignId($order->getPromoCampaignId());
			}
			else
			{
				// get data from order
				$promoData = array(
					'promo_type' => $order->getPromoType(),
					'discount_type' => $order->getPromoDiscountType(),
					'discount' => $order->getPromoDiscountValue(),
					'min_amount' => 0,
					'pid' => $order->getPromoCampaignId(),
				);
			}

			if (!is_null($promoData) && is_array($promoData))
			{
				$result = $this->calculatePromoDiscountAmount($order, $promoData, $subtotalAmount);

				if ($result) return $order->getPromoDiscountAmount();
			}
		}

		$this->reset($order);

		return 0.00;
	}

	public function checkPromoDiscount(ORDER $order, $promoCode)
	{
		$subtotalAmount = $order->calculateSubtotal(false, true);

		$promoData = $this->promoCodeRepository->getPromoCode($promoCode);

		if (!is_null($promoData) && is_array($promoData))
		{
			if ($this->calculatePromoDiscountAmount($order, $promoData, $subtotalAmount, false))
			{
				return $promoData;
			}
		}

		return null;
	}
	/**
	 * Get promo code discount amount
	 *
	 * @param ORDER $order
	 * @param $promoData
	 * @param $subtotalAmount
	 * @param $persist
	 *
	 * @return bool
	 */
	protected function calculatePromoDiscountAmount(ORDER $order, $promoData, $subtotalAmount, $persist = true)
	{
		switch ($promoData["promo_type"])
		{
			case 'Product': return $this->calculateProductBasedPromo($order, $promoData, $persist);
			case 'Shipping': return $this->calculateShippingBasedPromo($order, $promoData, $subtotalAmount, $persist);
			case 'Global': return $this->calculateGlobalBasedPromo($order, $promoData, $subtotalAmount, $persist);
		}

		return false;
	}

	/**
	 * Calculate global promo discount
	 *
	 * @param ORDER $order
	 * @param array $promoData
	 * @param float $subtotalAmount
	 * @param bool $persist
	 *
	 * @return bool
	 */
	protected function calculateGlobalBasedPromo(ORDER $order, $promoData, $subtotalAmount, $persist)
	{
		if ($subtotalAmount < $promoData["min_amount"])
		{
			return false;
		}

		$discountAmount = $this->calculateDiscountsForLineItems($order->getId(), $order->lineItems, $promoData, 1);

		if ($persist)
		{
			$order->setPromoDiscountAmount($discountAmount);

			$this->setOrderPromoData($order, $promoData['promo_type'], $promoData['pid'], $promoData['discount_type'], $promoData['discount'], $discountAmount);
		}

		return true;
	}

	/**
	 * Calculate shipping based promo discount
	 *
	 * @param ORDER $order
	 * @param array $promoData
	 * @param float $subtotalAmount
	 * @param bool $persist
	 *
	 * @return bool
	 */
	protected function calculateShippingBasedPromo(ORDER $order, $promoData, $subtotalAmount, $persist)
	{
		if ($subtotalAmount < $promoData['min_amount'])
		{
			return false;
		}

		$shippingAmount = $order->getShippingAmount() + $order->getShippingTaxAmount();

		if ($persist)
		{
			if ($promoData["discount_type"] == "amount")
			{
				$order->setPromoDiscountAmount(PriceHelper::round($promoData["discount"]));
			}
			else
			{
				$order->setPromoDiscountAmount(PriceHelper::round($shippingAmount / 100 * $promoData["discount"]));
			}

			if ($order->getPromoDiscountAmount() > $shippingAmount)
			{
				$order->setPromoDiscountAmount($shippingAmount);
			}

			$this->setOrderPromoData($order, $promoData['promo_type'], $promoData['pid'], $promoData['discount_type'], $promoData['discount'], $order->getPromoDiscountAmount());
		}

		return true;
	}

	/**
	 * Calculates product - based promo codes
	 *
	 * @param ORDER $order
	 * @param array $promoData
	 * @param bool $persist
	 *
	 * @return bool
	 */
	protected function calculateProductBasedPromo(ORDER $order, $promoData, $persist)
	{
		$promoProducts = $this->promoCodeRepository->getPromoProducts($order->getId(), $promoData['pid']);

		if (count($promoProducts) > 0)
		{
			$items = array();
			$subtotal = 0;
			$totalQuantity = 0;

			foreach ($promoProducts as $product)
			{
				foreach ($order->lineItems as $item)
				{
					/* @var $item Model_LineItem */
					if ($item->getPid() == $product['pid'] && $item->getIsGift() == 'No')
					{
						$items[] = $item;

						$totalQuantity += $item->getFinalQuantity();

						$subtotal += $item->getPriceWithTax() * $item->getFinalQuantity() - $item->getDiscountAmount();
					}
				}
			}

			if ($subtotal < $promoData['min_amount'])
			{
				return false;
			}

			$discountAmount = $this->calculateDiscountsForLineItems($order->getId(), $items, $promoData, $totalQuantity);

			if ($persist)
			{
				$order->setPromoDiscountAmount($discountAmount);

				$this->setOrderPromoData($order, $promoData['promo_type'], $promoData['pid'], $promoData['discount_type'], $promoData['discount'], $order->getPromoDiscountAmount());
			}

			return true;
		}

		return false;
	}

	/**
	 * Calculate discount for line items
	 *
	 * @param $orderId
	 * @param $items
	 * @param $promoData
	 *
	 * @return float|int
	 */
	protected function calculateDiscountsForLineItems($orderId, $items, &$promoData, $maxQuantity)
	{
		$totalQuantity = 0;
		$subtotal = 0;
		foreach ($items as $item)
		{
			/* @var $item Model_LineItem */
			$subtotal += $item->getSubtotal(true) - $item->getDiscountAmount();
			$totalQuantity += $item->getFinalQuantity();
		}

		if ($totalQuantity == $maxQuantity && $promoData['discount_type'] == 'amount')
		{
			$discountAmount = 0;
			foreach ($items as $item)
			{
				$discount = $promoData['discount'];
				if ($discount > $item->getFinalPrice())
				{
					$discount = $item->getFinalPrice();
				}
				$lineDiscountAmount = $discount * $item->getFinalQuantity();
				$item->setPromoDiscountAmount($lineDiscountAmount);
				$discountAmount += $lineDiscountAmount;

				$this->orderRepository->updateLineItemDiscount($orderId, $item->getId(), $item->getDiscountAmount(), $item->getPromoDiscountAmount());
			}
		}
		else
		{
			if ($totalQuantity > $maxQuantity)
			{
				$totalQuantity = $maxQuantity;
			}

			$dp = $this->getDiscountPercentage($subtotal, $promoData['discount_type'], $promoData['discount'] * $totalQuantity);

			$discountAmount = 0;
			foreach ($items as $item)
			{
				$lineSubTotal = $item->getSubtotal(true) - $item->getDiscountAmount();

				$lineDiscountAmount = PriceHelper::round($lineSubTotal * $dp / 100);
				$item->setPromoDiscountAmount($lineDiscountAmount);
				$discountAmount += $lineDiscountAmount;

				$this->orderRepository->updateLineItemDiscount($orderId, $item->getId(), $item->getDiscountAmount(), $item->getPromoDiscountAmount());
			}
		}

		return $discountAmount;
	}

	/**
	 * Get promo discount percentage
	 *
	 * @param $subtotalAmount
	 * @param $discountType
	 * @param $discountValue
	 *
	 * @return float|int
	 */
	protected function getDiscountPercentage($subtotalAmount, $discountType, $discountValue)
	{
		//determinate discount percent
		$dp = 0;

		if ($subtotalAmount > 0)
		{
			if ($discountType == 'percent')
			{
				$dp = $discountValue;
			}
			else if ($discountType == 'amount')
			{
				$dp = $discountValue / ($subtotalAmount / 100);
			}
		}

		return $dp > 100 ? 100 : $dp;
	}

	/**
	 * Sets order properties for promo code
	 *
	 * @param ORDER $order
	 * @param string $promoType
	 * @param int $promoCampaignId
	 * @param string $promoDiscountType
	 * @param float $promoDiscountValue
	 * @param float $promoDiscountAmount
	 */
	protected function setOrderPromoData(ORDER $order, $promoType, $promoCampaignId, $promoDiscountType, $promoDiscountValue, $promoDiscountAmount)
	{
		$order->setPromoType($promoType);
		$order->setPromoCampaignId($promoCampaignId);
		$order->setPromoDiscountType($promoDiscountType);
		$order->setPromoDiscountValue($promoDiscountValue);
		$order->setPromoDiscountAmount(PriceHelper::round($promoDiscountAmount));
	}

	/**
	 * Returns class instance
	 *
	 * @param $settings
	 * @param DataAccess_OrderRepositoryInterface $orderRepository
	 * @param DataAccess_PromoCodeRepository $promoCodeRepository
	 *
	 * @return Calculator_PromoDiscount
	 */
	// @codeCoverageIgnoreStart
	public static function getInstance(&$settings, DataAccess_OrderRepositoryInterface $orderRepository = null, DataAccess_PromoCodeRepository $promoCodeRepository = null)
	{
		if (is_null(self::$instance))
		{
			if (is_null($orderRepository))
			{
				global $db;
				$orderRepository = new DataAccess_OrderRepository($db);
			}

			if (is_null($promoCodeRepository))
			{
				global $db;
				$promoCodeRepository = new DataAccess_PromoCodeRepository($db);
			}

			self::$instance = new self($orderRepository, $promoCodeRepository, $settings);
		}

		return self::$instance;
	}

	/**
	 * Set instance
	 *
	 * @param Calculator_PromoDiscount $instance
	 */
	public static function setInstance(Calculator_PromoDiscount $instance)
	{
		self::$instance = $instance;
	}
	// @codeCoverageIgnoreEnd
}