<?php
/**
 * Calculate shipping
 */
class Calculator_Shipping
{
	protected static $instance;

	protected $settings;
	protected $shippingRepository;
	protected $orderRepository;
	protected $dobaOrderApi;
	protected $currencies;

	/**
	 * Class constructor
	 *
	 * @param array $settings
	 * @param DataAccess_ShippingRepositoryInterface $shippingRepository
	 * @param DataAccess_OrderRepositoryInterface $orderRepository
	 * @param DobaOrderApiInterface $dobaOrderApi
	 * @param CurrenciesInterface $currencies
	 */
	public function __construct(
		&$settings,
		DataAccess_ShippingRepositoryInterface $shippingRepository,
		DataAccess_OrderRepositoryInterface $orderRepository,
		DobaOrderApiInterface $dobaOrderApi,
		CurrenciesInterface $currencies
	)
	{
		$this->settings = $settings;
		$this->shippingRepository = $shippingRepository;
		$this->orderRepository = $orderRepository;
		$this->dobaOrderApi = $dobaOrderApi;
		$this->currencies = $currencies;
	}

	/**
	 * Reset order data
	 * 
	 * @param  ORDER  $order
	 */
	public function reset(ORDER $order)
	{
		// Reset line item shipping prices
		$this->orderRepository->resetLineItemShippingPrices($order->getId());

		$order->setShippingAmount(0);
	}

	/**
	 * Calculate
	 *
	 * @param DB $db
	 * @param $settings
	 * @param ORDER $order
	 * @param Shipping_MethodInterface $shippingMethod
	 * @param null $product_level_method
	 * @param $error_message
	 *
	 * @return float
	 */
	public function calculate(DB $db, &$settings, ORDER $order, Shipping_MethodInterface $shippingMethod, $product_level_method = null, &$error_message)
	{
		/**
		 * Init dollar-days
		 */
		$dollar_days_shipping_percent = 15;
		$dollar_days_shipping_min_amount = 6.95;

		/**
 		 * Select/calc data required for shipping price calculation
 		 */
		$shipping_price_at_product_level = 0;
		$shipping_amount_for_price_based = 0;
		$shipping_amount_for_price_based_discounted = 0;
		$shipping_amount_for_price_based_exists = false;

		$shipping_weight = 0;
		$shipping_items_count = 0;
		$shipping_price_dollar_days = 0;
		$shipping_amount_dollar_days = 0;
		$shipping_amount_doba = 0;
		$doba_items = array();
		$ship_user_products = false;

		$shippingAmount = 0;

		$shippingRepository = $this->shippingRepository;

		/**
		 * Check is there used product level method
		 */
		if (!is_null($product_level_method))
		{
			$result = $shippingRepository->getOrderItemsWithShippingPrice($order->getId(), $product_level_method["ssid"]);
		}
		else
		{
			$product_level_method = false;
			$result = $shippingRepository->getOrderItems($order->getId());
		}

		$order_content_shipping_price = array();

		if ($result) foreach ($result as $item)
		{
			if ($order->isDollarDaysProduct($item))
			{
				$shipping_price_dollar_days = $shipping_price_dollar_days + $item["admin_price"] * $item["admin_quantity"];
			}
			elseif ($order->isDobaProduct($item))
			{
				$doba_items[] = array("item_id"=>$item["product_id"], "quantity"=>$item["admin_quantity"]);
			}
			else
			{
				if (!$order->isItemShippable($item)) continue;

				if ($order->isProductLevelShipping($item))
				{
					if ($product_level_method)
					{
						$shipping_price_at_product_level = $shipping_price_at_product_level + $item["product_shipping_price"] * $item["admin_quantity"];
						$order_content_shipping_price[$item['ocid']] = array("is_shipping_price" => "Yes", "shipping_price" => $item["product_shipping_price"]);
					}
				}
				else
				{
					$shipping_items_count = $shipping_items_count + $item["admin_quantity"];
					$shipping_weight = $shipping_weight + $item["weight"] * $item["admin_quantity"];
					$shipping_amount_for_price_based = $shipping_amount_for_price_based + $item["admin_price"] * $item["admin_quantity"];
					$shipping_amount_for_price_based_exists = true;
				}
			}
		}

		$address_from = $shippingRepository->getShippingOriginAddress($order->getId(), $settings);

		/**
		 * Prepare to(destination) shipping address
		 */
		$addressData = $shippingRepository->getOrderShippingAddress($order->getId());
		$address_to = array(
			"address_type" => $addressData["shipping_address_type"],
			"name" => $addressData["shipping_name"],
			"address1" => $addressData["shipping_address1"],
			"address2" => $addressData["shipping_address2"],
			"city" => $addressData["shipping_city"],
			"state" => $addressData["state_name"],
			"state_abbr" => $addressData["state_abbr"],
			"state_code" => $addressData["state_abbr"],
			"country" => $addressData["country_name"],
			"country_iso_a2" => $addressData["iso_a2"],
			"country_iso_a3" => $addressData["iso_a3"],
			"country_iso_number" => $addressData["iso_number"],
			"zip" => $addressData["shipping_country"]=="1"?substr(trim($addressData["shipping_zip"]), 0, 5):trim($addressData["shipping_zip"])
		);

		/**
		 * Dollar Days
		 */
		if ($shipping_price_dollar_days > 0)
		{
			$shipping_amount_dollar_days = $shipping_price_dollar_days / 100 * $dollar_days_shipping_percent;
			if ($shipping_amount_dollar_days < $dollar_days_shipping_min_amount)
			{
				$shipping_amount_dollar_days = 	$dollar_days_shipping_min_amount;
			}

			$shipping_price_at_product_level = $shipping_price_at_product_level + $shipping_amount_dollar_days;
		}

		/**
		 * Doba
		 */
		if (count($doba_items) > 0)
		{
			$dobaOrderApi = $this->dobaOrderApi;

			$shipping_amount_doba = $dobaOrderApi->getShippingQuote($address_to, $doba_items);
			if ($shipping_amount_doba !== false)
			{
				$shipping_price_at_product_level = $shipping_price_at_product_level + $shipping_amount_doba;
			}
		}

		/**
		 * Update shipping_price in orders_content
		 */
		if ($product_level_method && count($order_content_shipping_price))
		{
			foreach ($order_content_shipping_price as $ocid=>$item)
			{
				$shippingRepository->updateOrderItemShippingPrice($ocid, $item["shipping_price"]);
			}
		}

		if ($shipping_amount_for_price_based_exists)
		{
			/**
			 * Check discounts
			 */
			$promo_discount_percent = $order->getPromoDiscountPercentage();

			$discount_percent = $order->getDiscountPercentage();

			$shipping_amount_for_price_based_discounted = 
				$shipping_amount_for_price_based - 
				$shipping_amount_for_price_based / 100 * $promo_discount_percent -
				$shipping_amount_for_price_based / 100 * $discount_percent;
		}
		else
		{
			$shipping_amount_for_price_based_discounted	= $shipping_amount_for_price_based;
		}

		if ($order->getShippingCarrierId() == "custom" && $order->getShippingMethodId() == "product_level")
		{
			$result_shipping_price = $shipping_price_at_product_level;
		}
		else
		{
			// try to calc
			$currencies = $this->currencies;
			$defaultCurrency = $currencies->getDefaultCurrency();
			$currencyExchangeRates = $currencies->getExchangeRates();

			$error_message = "";

			$result_shipping_price = $shippingMethod->calculate(
				$db, 
				$address_from,
				$address_to, 
				$shipping_weight, 
				$shipping_items_count,
				array(
					'before_discount' => $shipping_amount_for_price_based,
					'after_discount' => $shipping_amount_for_price_based_discounted
				), //price based
				$error_message,
				$defaultCurrency['code'],
				$currencyExchangeRates
			);

			$result_shipping_price += $shipping_price_at_product_level;

			if ($error_message != "")
			{
				$shippingAmount = 0.00;
				$order->setShippingCarrierId("custom");
				$order->setShippingMethodId("disabled");
			}
		}

		$shippingAmount = $result_shipping_price;

		$shippingAmount = PriceHelper::round($shippingAmount);

		$order->setShippingAmount($shippingAmount);

		/**
		 * Check for doba / dollar days orders
		 */
		if ($order->getShippingAmount() > 0 && $order->getShippingSSID() == "0")
		{
			$order->setShippingSSID(0);
			$order->setShippingCarrierId('custom');
			$order->setShippingMethodId('custom');
			$order->setShippingCMName('Shipping');
		}

		return $shippingAmount;
	}

	/**
	 * Returns class instance
	 *
	 * @param $settings
	 * @param DataAccess_ShippingRepositoryInterface $shippingRepository
	 * @param DataAccess_OrderRepositoryInterface $orderRepository
	 * @param DobaOrderApiInterface $dobaOrderApi
	 * @param CurrenciesInterface $currencies
	 *
	 * @return Calculator_Shipping
	 */
	// @codeCoverageIgnoreStart
	public static function getInstance(
		&$settings,
		DataAccess_ShippingRepositoryInterface $shippingRepository = null,
		DataAccess_OrderRepositoryInterface $orderRepository = null,
		DobaOrderApiInterface $dobaOrderApi = null,
		CurrenciesInterface $currencies = null
	)
	{
		if (is_null(self::$instance))
		{
			if (is_null($shippingRepository))
			{
				global $db;

				$shippingRepository = new DataAccess_ShippingRepository($db);
			}

			if (!$orderRepository)
			{
				$orderRepository = new DataAccess_OrderRepository();
			}

			if (is_null($dobaOrderApi))
			{
				global $db;

				require_once('content/classes/class.doba_order.php');
				$dobaOrderApi = new DobaOrderAPI($settings, $db);
			}

			if (is_null($currencies))
			{
				global $db;

				require_once('content/classes/class.currencies.php');
				$currencies = new Currencies($db);
			}

			self::$instance = new self($settings, $shippingRepository, $orderRepository, $dobaOrderApi, $currencies);
		}

		return self::$instance;
	}

	/**
	 * Set instance
	 *
	 * @param Calculator_Shipping $instance
	 */
	public static function setInstance(Calculator_Shipping $instance)
	{
		self::$instance = $instance;
	}
	// @codeCoverageIgnoreEnd
}